using System;
using UnityEngine;
using UnityEngine.EventSystems;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{
	public class Atomizer
	{
		MonoBehaviour core;
		private Queue<KeyValuePair<Atom, Action<Atom>>> atomCreatedQueue;
		private HashSet<string> hshRecentlyCreatedAtomUids;

		public Atomizer(MonoBehaviour script)
        {
			core = script;
			hshRecentlyCreatedAtomUids = new HashSet<string>();
		}

		public void Destroy()
		{
			if (atomCreatedQueue != null)
            {
				atomCreatedQueue.Clear();
				atomCreatedQueue = null;
			}
			if (hshRecentlyCreatedAtomUids != null)
			{
				hshRecentlyCreatedAtomUids.Clear();
				hshRecentlyCreatedAtomUids = null;
			}
			core = null;
		}

		/// <summary>
        /// Performs the specific action after the specified seconds using a Coroutine in Unity.
        /// </summary>
        /// <param name="waitForSeconds"></param>
        /// <param name="action"></param>
		public void WaitBeforeAction(float waitForSeconds, Action action)
        {
			core.StartCoroutine(CreateWaitBeforeAction(waitForSeconds, action));
		}

		private IEnumerator CreateWaitBeforeAction(float waitForSeconds, Action action)
        {
			yield return new WaitForSeconds(waitForSeconds);
			action();
		}

		/// <summary>
		/// Creates an atom async and runs the onAtomCreated function once created in the next update cycle (requires Atomizer.Update() to be run each update).
		/// </summary>
		/// <param name="atomType"></param>
		/// <param name="atomUid"></param>
		/// <param name="atomPostfixIfExist"></param>
		/// <param name="onAtomCreated"></param>
		public void CreateAtom(string atomType, string atomUid, string atomPostfixIfExist, Action<Atom> onAtomCreated)
		{
			core.StartCoroutine(Create(atomType, atomUid, atomPostfixIfExist, onAtomCreated));
		}

		/// <summary>
		/// Creates an atom async and runs the onAtomCreated function once created in the next update cycle (requires Atomizer.Update() to be run each update).
		/// </summary>
		/// <param name="atomType"></param>
		/// <param name="atomUid"></param>
		/// <param name="atomPostfixIfExist"></param>
		/// <param name="onAtomCreated"></param>
		private IEnumerator Create(string atomType, string atomuid, string atomPostfixIfExist, Action<Atom> onAtomCreated)
		{
			yield return new WaitForSeconds(0.2f);

			string atomId = atomuid;
			// Atom atom = SuperController.singleton.GetAtomByUid(atomId); // Old code checking if atom exist
			bool atomUidTaken = (SuperController.singleton.GetAtomByUid(atomId) != null || hshRecentlyCreatedAtomUids.Contains(atomId)); // new check
			if (atomUidTaken && atomPostfixIfExist != null && atomPostfixIfExist != "")
			{
				// If atom by uid already exist, add postfix
				atomId = atomuid + "_" + atomPostfixIfExist;
				// atom = SuperController.singleton.GetAtomByUid(atomId);
				atomUidTaken = (SuperController.singleton.GetAtomByUid(atomId) != null || hshRecentlyCreatedAtomUids.Contains(atomId));
			}
			// if it still exist, keep adding sequence nr to it! Max 99, then we recycle the last added atom over-and-over...
			string unsequencedAtomId = atomId;
			int sequence = 2;
			// while (atom != null && sequence < 99)
			while (atomUidTaken && sequence < 99)
			{
				atomId = (unsequencedAtomId + "#" + sequence.ToString());
				sequence++;
				// atom = SuperController.singleton.GetAtomByUid(atomId); 
				atomUidTaken = (SuperController.singleton.GetAtomByUid(atomId) != null || hshRecentlyCreatedAtomUids.Contains(atomId));
			}

			Atom atom = null;
			if (!atomUidTaken)
			{
				hshRecentlyCreatedAtomUids.Add(atomId);
				yield return SuperController.singleton.AddAtomByType(atomType, atomId);
				atom = SuperController.singleton.GetAtomByUid(atomId);
			}

			if (atom != null)
			{
				if (atomCreatedQueue == null)
                {
					atomCreatedQueue = new Queue<KeyValuePair<Atom, Action<Atom>>>();
				}
				atomCreatedQueue.Enqueue(new KeyValuePair<Atom, Action<Atom>>(atom, onAtomCreated));
			}
		}

		/// <summary>
		/// Handles any onAtomCreated Actions in the main update loop. Returns true if anything was consumed.
		/// </summary>
		public bool Update()
        {
			if (atomCreatedQueue != null && atomCreatedQueue.Count > 0)
            {
				KeyValuePair<Atom, Action<Atom>> p = atomCreatedQueue.Dequeue();
				if (!p.Key.destroyed)
				{
					p.Value(p.Key);
					return true;
				}
			} 
			else
            {
				hshRecentlyCreatedAtomUids.Clear();
			}
			return false;
        }

	}
}