using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{
    public class Path
    {
		/// <summary>
        /// Verifies and setup plugin paths and folders. If it doesn't work out it will return false and add some GUI text that the plugin can't start.
        /// </summary>
		public bool VerifyAndSetup(MVRScript script) 
		{
			Atom a = script.GetContainingAtom();
			if (a != null && (a.type == "SessionPluginManager" || a.type == "CoreControl"))
			{
				try
                {
					Init(script);
				} 
				catch (Exception e)
                {
					Path.CreateInitErrorElements(script, "Failed to read plugin path. Plugin can't start.", e);
					return false;
                }

				return VerifyLegacyPresetFolder(script) && VerifyPluginDataFolder(script);

			}
			SuperController.LogMessage("Plugin must be attached as a session plugin.");
			return false;
		}

		private static void CreateInitErrorElements(MVRScript script, string errorMsg, Exception e)
        {
			SuperController.LogError("");
			SuperController.LogError(errorMsg);

			// Create error-element in GUI:
			JSONStorableString jss = new JSONStorableString("cuabErrorTxt", errorMsg);
			UIDynamicTextField text = script.CreateTextField(jss);

			UIDynamicButton btn = script.CreateButton("See full exception");
			btn.button.onClick.AddListener(() => {
				SuperController.singleton.ClearErrors();
				SuperController.LogError("");
				SuperController.LogError(errorMsg);
				SuperController.LogError("");
				SuperController.LogError(e.ToString());
				SuperController.singleton.OpenErrorLogPanel();
			});
		}

		/// <summary>
        /// Init plugin path and preset path for the script.
        /// </summary>
		public void Init(MVRScript script)
        {
			SetPluginPath(script);
			SetPluginPresetPath(script);
		}

		public string PluginPath { get { return cachedPluginPath; } }

		public string PluginPresetPath { get { return cachedPluginPresetPath; } }

		/// <summary>
        /// VAM 1.20.77 new plugin data folder
        /// </summary>
		public string PluginDataFolder { get { return "Saves\\PluginData\\Blazedust\\CUAManager"; } }

		private string cachedPluginPresetPath = "";
		private string SetPluginPresetPath(MVRScript script)
		{
			if (cachedPluginPresetPath != "")
			{
				return cachedPluginPresetPath;
			}
			SetPluginPath(script);
			cachedPluginPresetPath = Path.Combine(cachedPluginPath, "presets");
			return cachedPluginPresetPath;
		}

		private string cachedPluginPath = "";
		private string SetPluginPath(MVRScript script)
		{
			if (cachedPluginPath != "")
			{
				return cachedPluginPath;
			}

			string loadDir = SuperController.singleton.currentLoadDir;
			// Log(SuperController.singleton.currentSaveDir);
			// Log(script.storeId);
			string pluginId = script.storeId.Split('_')[0]; // first part is the plugin ID in the MVRPluginManager.
			// MVRPluginManager manager = containingAtom.GetStorableByID("PluginManager") as MVRPluginManager; // this is needed to find the manager on non-session plugins.
			MVRPluginManager man = (script.manager != null ? script.manager : (script.containingAtom.GetStorableByID("PluginManager") as MVRPluginManager)); // Not tested on a non-session plugin, but should work!
			string pathToScriptFile = man.GetJSON(true, true)["plugins"][pluginId].Value;
			// SuperController.LogError(pathToScriptFile); // "./Scripts/FolderA/PluginFolder/ADD_ME.cslist"
			string pathToScriptFolder = "";
			if (pathToScriptFile.Contains("/") || pathToScriptFile.Contains("\\")) { pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' })); }
			if (pathToScriptFolder.StartsWith(".")) { pathToScriptFolder = pathToScriptFolder.Substring(1); }
			if (pathToScriptFolder.StartsWith("/")) { pathToScriptFolder = pathToScriptFolder.Substring(1); }
			// SuperController.LogError(pathToScriptFolder); // "Scripts/FolderA/PluginFolder" OR "AddonPackageVarFile.#:/Scripts/FolderA/PluginFolder" for var-packages.
			// Plugin path should be a non-varpackage container (don't write back to var-packages, let preset files be in the folder structure, and for backwards compatibility for older versions of this plugin!)
			// Failsafe: We only cut the first "/" if it is preceded by a ":" character, assuming var packages are always split with the ":/" separator.
			int varPackageIndex = pathToScriptFolder.IndexOf("/");
			if (varPackageIndex > 0 && pathToScriptFolder.IndexOf(":/") == (varPackageIndex - 1))
			{
				pathToScriptFolder = pathToScriptFolder.Substring(varPackageIndex + 1);
			}
			if (!pathToScriptFolder.StartsWith("Custom/Scripts/"))
			{
				// Assuming the script is in the same folder as the scene file OR in a subfolder of the scene file when the path doesn't start with "Custom/Scripts/".
				if (pathToScriptFolder != "")
				{
					if (loadDir != "")
					{
						pathToScriptFolder = loadDir + "/" + pathToScriptFolder;
					}
				}
				else
				{
					pathToScriptFolder = loadDir;
				}
			}
			// SuperController.LogError(pathToScriptFolder); // "Custom/Scripts/FolderA/PluginFolder" OR  "Saves/scene/User/SceneName/Scripts/FolderA/PluginFolder"
			cachedPluginPath = pathToScriptFolder;
			return cachedPluginPath;
		}

		public PresetFileInfo GetSavePresetFileInfo(string presetName)
		{
			if (presetName == null || presetName == "")
			{
				return new PresetFileInfo("", "", "", false, false);
			}
			presetName = presetName.Trim();
			string fileName = Path.GetFileNameFromPresetName(presetName);
			string folder = PluginPresetPath;
			// Log("Folder: " + folder);
			string[] existingFiles = SuperController.singleton.GetFilesAtPath(folder, "*.json");
			string filePath = Path.Combine(folder, fileName);
			filePath = filePath.Replace("\\", "/"); // Sanitize input, just-in-case.
			bool fileExist = (existingFiles != null && existingFiles.Where((x) => {
				// Log("Comparing " + x + " with " + filePath);
				// GetFilesAtPath() will return a mix of "/" and "\". Sanitize input here for x.
				return string.Equals(x.Replace("\\", "/"), filePath, StringComparison.InvariantCultureIgnoreCase);
			}).Any());

			return new PresetFileInfo(presetName, fileName, filePath, fileExist, true);
		}

		public string GetPresetNameFromFileSelection(string fileSelection)
        {
			if (loadPresetSelectionNameToFile != null && fileSelection != "")
			{
				// VAM's GetFilesAtPath() will return a mix of "/" and "\" but other parts will return only "/" ... messy!
				foreach (KeyValuePair<string, string> pair in loadPresetSelectionNameToFile)
                {
					// SuperController.LogMessage(pair.Key + " => " + pair.Value);
					if (pair.Value.Replace("\\", "/") == fileSelection.Replace("\\", "/"))
                    {
						return pair.Key;
                    }
                }
			}
			return "";
		}

		public string GetFileFromPresetSelection(string presetSelection)
		{
			string file = "";
			if (loadPresetSelectionNameToFile != null && presetSelection != "" && loadPresetSelectionNameToFile.TryGetValue(presetSelection, out file))
			{
				return file;
			}
			return "";
		}

		private Dictionary<string, string> loadPresetSelectionNameToFile;
		private List<string> loadPresetSelectionNames;
		public List<string> GetLoadPresetSelectionNames(bool clearCache = false)
		{
			// Return a cached list, until "Refresh" button is pressed.
			if (loadPresetSelectionNames == null || clearCache)
			{
				loadPresetSelectionNames = new List<string>();
				loadPresetSelectionNameToFile = new Dictionary<string, string>();
				string[] files = SuperController.singleton.GetFilesAtPath(PluginPresetPath, "*.json");
				// Could spread out this load over several updates or in a coroutine, but let's just freeze application until it's done. Good enough! Simplifies how the GUI is updated too!
				Array.Sort(files, StringComparer.InvariantCulture);
				for (int i = 0; i < files.Length; i++)
				{
					string file = files[i];
					// assuming the files are all valid plugin-json formatted in here! This builds the list even faster.
					// The filename will also be the key for selected entry.
					// File is in format "Preset_*.json". Only want the * part as the name.
					string fileName = Path.GetFileNameAndExt(file);

					//// do NOT include dummy.json file in case it ever was created from plugin installation and do NOT include option files...
					//if (string.Compare(fileName, "dummy.json", true) == 0 || fileName.StartsWith("options_", StringComparison.InvariantCultureIgnoreCase))
					//{
					//	continue;
					//}

					// ... actually only include files beginning with "preset_"
					if (!fileName.StartsWith("preset_", StringComparison.InvariantCultureIgnoreCase))
                    {
						continue;
                    }

					string presetName = Path.GetPresetNameFromFileName(fileName);
					loadPresetSelectionNames.Add(presetName);
					loadPresetSelectionNameToFile[presetName] = file;
				}
				// lastly, insert an empty selection
				loadPresetSelectionNames.Insert(0, ""); 
			}
			return loadPresetSelectionNames;
		}

		public bool VerifyLegacyPresetFolder(MVRScript script)
        {
			// Verify plugin preset path
			string[] existingFiles = null;
			try
			{
				existingFiles = SuperController.singleton.GetFilesAtPath(PluginPresetPath, "*.json");
			}
			catch (Exception e)
			{
				SuperController.singleton.ClearErrors();
				if (e.GetType().FullName == "System.IO.DirectoryNotFoundException" || e.Message != null && e.Message.Contains("non-existent path"))
				{
					SuperController.LogMessage("Missing folder '" + PluginPresetPath + "'. Attempting to automatically create it.");
					if (Path.TryToCreatePluginFolder(PluginPresetPath))
					{
						return true;
					}
				}
				Path.CreateInitErrorElements(script, "Looks like the folder '" + PluginPresetPath + "' is missing. Plugin can't start without it. Create the folder and reload the plugin.", e);
				return false;
			}
			return true;
		}

		public bool VerifyPluginDataFolder(MVRScript script)
		{
			// FileManagerSecure.DirectoryExists(PluginDataFolder)
			// FileManagerSecure.CreateDirectory(PluginDataFolder)
			string[] existingFiles;
			try
			{
				existingFiles = SuperController.singleton.GetFilesAtPath(PluginDataFolder, "*.json");
			}
			catch (Exception e)
			{
				SuperController.singleton.ClearErrors();
				if (e.GetType().FullName == "System.IO.DirectoryNotFoundException" || e.Message != null && e.Message.Contains("non-existent path"))
				{
					SuperController.LogMessage("Missing folder '" + PluginDataFolder + "'. Attempting to automatically create it.");
					if (Path.TryToCreatePluginFolder(PluginDataFolder))
					{
						return true;
					}
				}
				Path.CreateInitErrorElements(script, "Looks like the folder '" + PluginDataFolder + "' is missing. Plugin can't start without it. Create the folder and reload the plugin.", e);
				return false;
			}
			return true;
		}

		public static bool TryToCreatePluginFolder(string presetPath)
		{
			// Attempt to create the folder. Do this by saving a dummy-file
			Camera cam = SuperController.singleton.screenshotCamera;
			try
			{

				string nPath = presetPath + "/dummy.json";
				// For the dummy-save to automatically create the directory, the path requires backslashes! A little quirk.
				nPath = nPath.Replace("/", "\\");

				// disable screenshot taking 
				// don't do this if you don't know what you're doing! 
				// This might not be allowed in future versions of VaM
				// we do restore it in the end of this function though!
				SuperController.singleton.screenshotCamera = null;

				// Save a dummy file only containing one atom to keep it small and speedy.
				Atom prioAtom = Helpers.GetFirstAtomOfType(new string[3] { "WindowCamera", "PlayerNavigationPanel", "InvisibleLight" });

				SuperController.singleton.Save(nPath, prioAtom, includePhysical: false, includeAppearance: false);

				// check files again!
				try
				{
					string[] existingFiles = SuperController.singleton.GetFilesAtPath(presetPath, "*.json");

					// Can read files now, assuming folder was created!
					SuperController.LogMessage("Creating folder successful!");

					// plugin folder created!
					return true;
				}
				catch (Exception exce)
				{
					// Still no luck!
				}
			}
			catch (Exception ex)
			{
				// Even failed to create the directory with a dummy file... User must create this folder!
				SuperController.LogMessage("Automatic creation failed");
				SuperController.LogMessage(ex.ToString());
			}
			finally
			{
				SuperController.singleton.screenshotCamera = cam;
			}

			return false;
		}

		public static string Combine(string a, string b)
		{
			return a + (a.EndsWith("/") ? "" : "/") + b;
		}
		public static string GetFileNameAndExt(string filePath)
		{
			int index = Math.Max(filePath.LastIndexOf("/"), filePath.LastIndexOf("\\"));  // also checking "\\"
			if (index >= 0)
			{
				filePath = filePath.Substring(index + 1);
			}
			return filePath;
		}

		// This does not sanitize any input. The output name can be invalid, but we ignore that as the user can see exceptiopn in the log if they happen.
		public static string GetFileNameFromPresetName(string presetName)
		{
			return "preset_" + presetName + ".json";
		}
		public static string GetPresetNameFromFileName(string fileName)
		{
			string presetName = fileName;
			if (presetName.ToLower().StartsWith("preset_"))
			{
				presetName = presetName.Substring("preset_".Length);
			}
			if (presetName.ToLower().EndsWith(".json"))
			{
				presetName = presetName.Substring(0, presetName.Length - ".json".Length);
			}
			return presetName;
		}

	}

	public class PresetFileInfo
	{
		public string presetName; // x
		public string fileName; // preset_x.json
		public string filePath; // custom/scripts/.../preset_x.json
		public bool fileExist; // if file already exist on disk
		public bool isValid; // if data is valid

		public PresetFileInfo(string presetName, string fileName, string filePath, bool fileExist, bool isValid)
		{
			this.presetName = presetName;
			this.fileName = fileName;
			this.filePath = filePath;
			this.fileExist = fileExist;
			this.isValid = isValid;
		}
	}

}