using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MeshedVR;

namespace Blazedust.CUAM
{

    

    public class PluginHelper
    {
        protected Queue<PluginQueueItem> _queue;
        protected byte _nextUpdate = 0;

        public bool Verbose { get { return core.options.verbosePluginLogs.val; } }

        public CustomUnityAssetManager core;
        public PluginHelper(CustomUnityAssetManager cuam)
        {
            this.core = cuam;
            _queue = new Queue<PluginQueueItem>();
        }
        
        public void QueueLoadPluginPreset(Atom rootAtom, AutoLoadTriggerInfo triggerInfo)
        {
            // SuperController.LogMessage("Load " + triggerInfo.presetName);
            _queue.Enqueue(new PluginQueueItem(rootAtom, true, triggerInfo));
        }

        public void QueueRemovePluginPreset(Atom rootAtom, AutoLoadTriggerInfo triggerInfo)
        {
            // SuperController.LogMessage("Remove " + triggerInfo.presetName);
            _queue.Enqueue(new PluginQueueItem(rootAtom, false, triggerInfo));
        }

        public void Update()
        {
            // any queued things to process?
            if (_queue.Count == 0)
                return;

            if (Helpers.IsLoading())
            {
                if (_nextUpdate < 40)
                    _nextUpdate = 40;
                return;
            }

            // only process one item per 20 update cycles, it could be more agressive but let any plugins settle and do their init job whatever it could be.
            _nextUpdate--;
            if (_nextUpdate > 0)
                return;

            _nextUpdate = 20;
            PluginQueueItem p = _queue.Dequeue();
            if (p.RootAtom == null || p.RootAtom.destroyed)
                return;

            if (p.Mode)
                LoadPluginTrigger(p.RootAtom, p.TriggerInfo);
            else
                RemovePluginTrigger(p.RootAtom, p.TriggerInfo);
        }

        public void LoadPluginTrigger(Atom rootAtom, AutoLoadTriggerInfo triggerInfo)
        {
            if (Verbose)
                SuperController.LogMessage("CUAM: Trigger loading plugin " + triggerInfo.presetName);

            PluginPresetInfo presetInfo = GetPluginPresetInfo(triggerInfo.presetName);
            if (!presetInfo.IsValid)
            {
                if (Verbose)
                    SuperController.LogMessage("CUAM: Plugin preset '" + triggerInfo.presetName + "' not found from trigger.");
                return;
            }

            string presetHash = GetPluginPresetHash(triggerInfo.presetName);

            if (Verbose)
                SuperController.LogMessage("CUAM: Target hash: " + presetHash);

            bool pluginLoaded = false;

            PluginManagerInfo pmi = GetPluginManagerInfo(rootAtom);
            try
            {
                // Failsafe 1 - Check plugin trigger already loaded
                pluginLoaded = pmi.Plugins.Any((x) =>
                {
                    return x.Scripts.Any((s) => { 
                        if (PluginHelper.PluginHashSameSource(s.scriptLabel, presetHash))
                        {
                            if (Verbose)
                                SuperController.LogMessage("CUAM: Target hash already found among plugins: " + s.scriptUid + " with hash " + s.scriptLabel);

                            return true;
                        }
                        return false;
                    });
                });
                if (pluginLoaded)
                {
                    if (Verbose)
                        SuperController.LogMessage("CUAM: Skipping plugin preset '" + triggerInfo.presetName + "', already loaded from trigger.");
                    return;
                }

                // Failsafe 2, skip loading plugin preset if ALL existing plugins match perfectly. 
                // We assume this is from the original scene from which the plugin preset was created, if user is creating a plugin!
                if (presetInfo.PluginURLs.Count == pmi.Plugins.Count)
                {
                    // potential to match
                    List<string> existingPluginURLs = new List<string>();
                    foreach (PluginInfo pi in pmi.Plugins.Where((x) => { return !x.Scripts.Any((s) => { return IsCUAMPluginHashSource(s.scriptLabel); }); }))
                    {
                        existingPluginURLs.Add(pi.pluginURL);
                    }

                    // existingPluginURLs still match the preset info plugin count.
                    if (presetInfo.PluginURLs.Count == existingPluginURLs.Count)
                    {
                        foreach (string presetPluginUrl in presetInfo.PluginURLs)
                        {
                            existingPluginURLs.Remove(presetPluginUrl);
                        }

                        if (existingPluginURLs.Count == 0)
                        {
                            // all plugins accounted for with exact version matching. Ignore!
                            if (Verbose)
                                SuperController.LogMessage("CUAM: Skipping plugin preset '" + triggerInfo.presetName + "', already fully loaded.");
                            return;
                        }
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogMessage("WTF");
                SuperController.LogMessage(e.ToString());
            }
            finally
            {
                pmi.Dispose();
            }

            // Go ahead and load
            // This will "collapse" the pluginUid sequence, starting from 0 again.
            // The plugin preset is added last. We use this to grab the plugins and update the pluginLabels with "CUAM_hash".
            core.LoadPreset(rootAtom, "PluginPresets", "Custom\\Atom\\Person\\Plugins", triggerInfo.presetName, true, () =>
            {
                try
                {
                    // Grab the last added plugins in the list and mark them with the hashes.
                    pmi = GetPluginManagerInfo(rootAtom);
                    int pluginCount = presetInfo.PluginURLs.Count;
                    int startIndex = pmi.Plugins.Count - pluginCount;
                    if (startIndex < 0) // happens if pluginPreset fails to load for some reason, failsafe
                        startIndex = 0;
                    for (int i = startIndex; i < pmi.Plugins.Count; i++)
                    {
                        foreach (ScriptInfo si in pmi.Plugins[i].Scripts)
                        {
                            if (si.scriptLabel != "" && Verbose)
                                SuperController.LogMessage("CUAM: Changing existing scriptLabel " + si.scriptLabel);

                            si.scriptLabel = GetPluginPresetHash(triggerInfo.presetName);
                        }
                    }
                } 
                catch (Exception e)
                {
                    SuperController.LogError("CUAM: Failed to set pluginLabels");
                    SuperController.LogError(e.ToString());
                }
            });
        }

        public void RemovePluginTrigger(Atom rootAtom, AutoLoadTriggerInfo triggerInfo)
        {
            if (Verbose)
                SuperController.LogMessage("CUAM: Trigger removing plugin " + triggerInfo.presetName);

            string presetHash = GetPluginPresetHash(triggerInfo.presetName);

            bool pluginRemoved = true;
            while (pluginRemoved)
            {
                pluginRemoved = false;
                PluginManagerInfo pmi = GetPluginManagerInfo(rootAtom);
                try
                {
                    if (pmi.Plugins.Count == 0)
                        return;

                    bool anyRemoved = false;
                    foreach (PluginInfo pi in pmi.Plugins)
                    {
                        if (pi.Scripts.Any((s) => { return PluginHelper.PluginHashSameSource(s.scriptLabel, presetHash); }))
                        {
                            anyRemoved = true;
                            // this plugin is to be removed!
                            try
                            {
                                pmi.PluginManager.RemovePluginWithUID(pi.pluginUid);
                            }
                            catch (Exception e)
                            {
                                SuperController.LogError(e.ToString());
                                SuperController.LogMessage("CUAM: Failed to remove plugin " + pi.pluginUid);
                                return;
                            }
                            // as the plugin uids can collapse, rescan plugins and remove until all plugins linked to the preset are removed.
                            pluginRemoved = true;
                            break;
                        }
                    }
                    if (Verbose && !anyRemoved)
                    {
                        SuperController.LogMessage("CUAM: No plugin found with matching source to remove. The plugin might have been removed already, reloaded or label updated from other sources which CUAM does not support.");
                    }
                }
                finally
                {
                    pmi.Dispose();
                }

                if (!pluginRemoved)
                    break;
            }
        }

        public void Destroy()
        {
            this.core = null;
        }

        /// <summary>
        /// Any plugins in the pluginPreset with the same pluginUrl already loaded on the person will be cleared.
        /// </summary>
        /// <param name="allowDifferentVarVersions">If var package versions may differ</param>
        public void ClearPluginsFromPreset(Atom rootAtom, string pluginPreset, bool allowDifferentVarVersions = true)
        {
            PluginPresetInfo presetInfo = GetPluginPresetInfo(pluginPreset);
            if (!presetInfo.IsValid)
                return;

            PluginManagerInfo pmi = GetPluginManagerInfo(rootAtom);
            // presetInfo.PluginURLs
            // Match plugin URLs with current loaded plugin uids in pmi. Remove those!
            try
            {
                List<string> pluginUidsToClear = new List<string>();
                foreach(string presetPluginURL in presetInfo.PluginURLs)
                {
                    foreach(PluginInfo pi in pmi.Plugins.Where((x) => {
                        return PluginURLMatch(x.pluginURL, presetPluginURL, allowDifferentVarVersions) && !x.Scripts.Any((s) => IsCUAMPluginHashSource(s.scriptLabel));
                    }))
                    {
                        pluginUidsToClear.Add(pi.pluginUid);
                    }
                }
                foreach(string pluginUid in pluginUidsToClear)
                {
                    try
                    {
                        pmi.ClearPluginUid(pluginUid, Verbose);
                    } 
                    catch (Exception e)
                    {
                        SuperController.LogError(e.ToString());
                        SuperController.LogMessage("CUAM: Failed to clear plugin " + pluginUid);
                    }
                }
            } 
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
            pmi.Dispose();
        }

        public static bool PluginURLMatch(string pluginUrlA, string pluginUrlB, bool allowDifferentVarVersions = false)
        {
            if (pluginUrlA == null || pluginUrlB == null)
                return false;

            if (allowDifferentVarVersions)
            {
                // Author.Plugin.X:/InternalPath/SubFolder/File.cslist|cs|dll*
                // Author.Plugin.X:/InternalPath/SubFolder/File.cslist|cs|dll*
                // Remove the characters just before first occurance of ":/" and validate file from that.
                return StripPluginVarVersion(pluginUrlA) == StripPluginVarVersion(pluginUrlB);
            }
            return pluginUrlA == pluginUrlB;
        }

        public static string StripPluginVarVersion(string pluginUrl)
        {
            // Remove the characters just before first occurance of ":/" and validate file from that.
            int index = pluginUrl.IndexOf(":/");
            if (index == -1)
                return pluginUrl; // not a var format

            string pluginUrlPartA = pluginUrl.Substring(0, index);
            string pluginUrlPartB = pluginUrl.Substring(index);

            index = pluginUrlPartA.LastIndexOf(".");
            if (index >= 0) 
            {
                pluginUrlPartA = pluginUrlPartA.Substring(0, index) + ".x";
            }

            string strippedUrl = pluginUrlPartA + pluginUrlPartB;
            // SuperController.LogMessage("Stripped var url: " + pluginUrl + " => " + strippedUrl);
            return strippedUrl;
        }

        public PluginManagerInfo GetPluginManagerInfo(Atom rootAtom)
        {
            PluginManagerInfo pmi = new PluginManagerInfo(rootAtom);
            pmi.ReadCurrentPluginStatus(Verbose);
            return pmi;
        }

        public static PluginPresetInfo GetPluginPresetInfo(string pluginPreset)
        {
            PluginPresetInfo info = new PluginPresetInfo();
            info.IsValid = false;
            info.PluginPreset = pluginPreset;
            info.PluginURLs = new List<string>();

            try
            {
                JSONNode jsonNode = SuperController.singleton.LoadJSON(pluginPreset);
                if (jsonNode == null)
                {
                    info.IsValid = false;
                    return info;
                }

                info.IsValid = true;

                if (jsonNode["storables"] == null)
                    return info;

                JSONNode pluginManagerNode = jsonNode["storables"].AsArray.Childs.FirstOrDefault((x) => {
                    return x["id"].Value == "PluginManager";
                });
                
                if (pluginManagerNode == null)
                    return info;

                JSONNode jsonNodePlugins = pluginManagerNode["plugins"];
                if (jsonNodePlugins == null)
                    return info;

                JSONClass jc = jsonNodePlugins.AsObject;
                foreach (string key in jc.Keys)
                {
                    // Same plugin can be added multiple times here with different plugin uids.
                    info.PluginURLs.Add(jc[key]);
                }
            }
            catch (Exception e)
            {
                info.IsValid = false;
                SuperController.LogError(e.ToString());
            }

            return info;
        }

        public static void Test(Atom rootAtom)
        {
            //JSONStorable store = rootAtom.GetStorableByID("PluginManager");
            //if (store != null)
            //{
            //    SuperController.LogMessage("Has a store");
            //    JSONStorable store2 = store.GetStorableByID("plugin#0"); // N/A
            //    if (store2 != null)
            //    {
            //        SuperController.LogMessage("plugin#0 found");
            //        return;
            //    }
            //    SuperController.LogMessage("But no cigarr");
            //    return;
            //}
            //SuperController.LogMessage("Nope");

            // If it's possible to load plugin preset .json and grab plugin data nothing needs to be stored.
            // Everything can then be read live from whatever data is already existing on the atom.
            // PluginPresetInfo info = GetPluginPresetInfo("Custom/Atom/Person/Plugins/Preset_Glance+Expression.vap");
            //SuperController.LogMessage(info.PluginPreset);
            //foreach (string s in info.PluginURLs)
            //{
            //    SuperController.LogMessage(s);
            //}

            //string pluginPreset = "Custom/Atom/Person/Plugins/Preset_Glance+Expression.vap";
            //string hashA = GetPluginPresetHash(pluginPreset);
            //string hashB = GetPluginPresetHash(pluginPreset);
            //SuperController.LogMessage(hashA);
            //SuperController.LogMessage(hashB);
            //SuperController.LogMessage("Same=" + PluginHashSameSource(hashA, hashB).ToString());
        }

        public static bool IsCUAMPluginHashSource(string pluginLabel)
        {
            return pluginLabel.StartsWith("CUAM_");
        }

        public static bool PluginHashSameSource(string pluginPresetA, string pluginPresetB)
        {
            if (pluginPresetA == null || pluginPresetB == null || pluginPresetA == "" || pluginPresetB == "")
                return false;

            if (!pluginPresetA.StartsWith("CUAM_") || !pluginPresetB.StartsWith("CUAM_"))
                return false;

            string presetHashA = pluginPresetA;
            {
                int indexA = pluginPresetA.IndexOf("_", 5);
                if (indexA > 0)
                    presetHashA = pluginPresetA.Substring(indexA + 1);
            }

            // SuperController.LogMessage(presetHashA);

            string presetHashB = pluginPresetB;
            {
                int indexB = pluginPresetB.IndexOf("_", 5);
                if (indexB > 0)
                    presetHashB = pluginPresetB.Substring(indexB + 1);
            }

            // SuperController.LogMessage(presetHashB);

            return presetHashA == presetHashB;
        }

        /// <summary>
        /// Input: "Custom/Atom/Person/Plugins/Preset_Glance+Expression.vap" <br/>
        /// Output: CUAM_iVnE3AmzP_700gHw#Glance+Expression <br/>
        /// "CUAM" + _ + "iVnE3AmzP" + "_" + "700gHw#Glance+Expression" <br/>
        /// Prefix, fixed. "CUAM" <br/>
        /// Unique ID during the same session of VAM when plugin create plugins, based on ticks and an internal counter: "iVnE3AmzP" <br/>
        /// Postfix, fixed hash of the plugin string + plugin preset name "700gHw#Glance+Expression". 
        /// In theory this could cause ecollisions between plugin presets with the same name within different folders, but the local string hash containing the path is good enough to avoid collisions!
        /// </summary>
        public static string GetPluginPresetHash(string pluginPreset)
        {
            int pluginHash = pluginPreset.GetHashCode();
            string base64 = Convert.ToBase64String(BitConverter.GetBytes(pluginHash)).Replace("=", "").Replace("/", "-");

            int i = pluginPreset.LastIndexOf("/Preset_");
            if (i == -1)
                pluginPreset.LastIndexOf("\\Preset_");
            if (i > 0)
            {
                string name = pluginPreset.Substring(i + "/Preset_".Length);
                if (name.EndsWith(".vap", StringComparison.InvariantCultureIgnoreCase))
                    name = name.Substring(0, name.Length - 4);


                base64 = base64 + "#" + name;
            } 
            else
            {
                base64 = base64 + "#na";
            }

            string guidish = ShortGuidish();

            return "CUAM_" + guidish + "_" + base64;
        }

        private static long _sequence = 1;
        /// <summary>
        /// This generates same values if generating more than 100 results per TICK! CUAM is not designed to do that anyway!
        /// </summary>
        public static string ShortGuidish()
        {
            long t = DateTime.Now.Ticks + (_sequence++);
            int hshA = ((int)t).GetHashCode();
            int hshB = ((int)(t >> 32)).GetHashCode();
            string base64 = Convert.ToBase64String(BitConverter.GetBytes(hshA)).Replace("=", "").Replace("/", "-");
            base64 += Convert.ToBase64String(BitConverter.GetBytes(hshB)).Replace("=", "").Replace("/", "-").Substring(0, 3); // the last chars will never change on the same session of VAM, cut to save space!
            return base64;
        }

        //public static void Test(Atom rootAtom)
        //{
        //    if (rootAtom == null)
        //    {
        //        SuperController.LogMessage($"{nameof(rootAtom)} == null");
        //        return;
        //    }

        //    //foreach(string storableId in rootAtom.GetStorableIDs())
        //    //{
        //    //    bool storeIdReported = false;
        //    //    JSONStorable store = rootAtom.GetStorableByID(storableId);
        //    //    foreach (string fkey in store.GetFloatParamNames())
        //    //    {
        //    //        if (fkey.ToLowerInvariant().Contains("scale"))
        //    //        {
        //    //            if (!storeIdReported)
        //    //            {
        //    //                SuperController.LogMessage($"storableId: {storableId}");
        //    //                storeIdReported = true;
        //    //            }
        //    //            SuperController.LogMessage($"{fkey} = {store.GetFloatParamValue(fkey)}");
        //    //        }
        //    //    }
        //    //}

        //    MVRPluginManager pluginManager = rootAtom.GetStorableByID("PluginManager") as MVRPluginManager;
        //    if (pluginManager == null)
        //    {
        //        SuperController.LogMessage($"{nameof(pluginManager)} == null");
        //        return;
        //    }

        //    // list existing plugins:
        //    JSONClass json = pluginManager.GetJSON(true, true, true);
        //    if (json == null)
        //        SuperController.LogMessage($"{nameof(json)} == null");

        //    SuperController.LogMessage(json.ToString());

        //    JSONNode jsonNodePlugins = json["plugins"];
        //    JSONClass jsonClassPlugins = jsonNodePlugins.AsObject;

        //    List<string> scriptKeys = new List<string>();
        //    foreach (string storeableId in rootAtom.GetStorableIDs())
        //    {
        //        if (storeableId.StartsWith("plugin#"))
        //        {
        //            scriptKeys.Add(storeableId);
        //        }
        //    }

        //    foreach (string key in jsonClassPlugins.Keys)
        //    {
        //        SuperController.LogMessage($"{key} = {jsonClassPlugins[key].Value}"); // plugin#0 | plugin#1 | plugin#2 ... list uid!

        //        // find script storables starting with the plugin key, read it's label
        //        foreach(string scriptKey in scriptKeys)
        //        {
        //            if (scriptKey.StartsWith(key + "_"))
        //            {
        //                //try
        //                //{
        //                //    MVRPlugin plugin = pluginManager.GetStorableByID(key) as MVRPlugin;
        //                //    SuperController.LogMessage(plugin.ToString());
        //                //} 
        //                //catch (Exception e)
        //                //{
        //                //    SuperController.LogMessage(e.ToString());
        //                //}

        //                SuperController.LogMessage(scriptKey);
        //                // JSONStorable store = rootAtom.GetStorableByID(scriptKey);
        //                // SuperController.LogMessage(store.ToString());
        //                MVRScript script = rootAtom.GetStorableByID(scriptKey) as MVRScript;
        //                SuperController.LogMessage(script.pluginLabelJSON.val);
        //                // string base64Guidish = Convert.ToBase64String(Guid.NewGuid().ToByteArray()).Replace("=", "").Replace("/", "-");
        //                // script.pluginLabelJSON.val = "CUAM_" + base64Guidish;
        //            }
        //        }
                
        //        // pluginManager.RemovePluginWithUID(key); // This removes the plugin
        //        //try
        //        //{
        //        //    MVRScript script = rootAtom.GetStorableByID(key) as MVRScript;
        //        //    SuperController.LogMessage(script.pluginLabelJSON.val);
        //        //} 
        //        //catch (Exception e)
        //        //{
        //        //    SuperController.LogMessage(e.ToString());
        //        //}
        //    }

        //    // Now we know which uid:s exist BEFORE

        //    // Then we can know which uid:s exist AFTER

        //    // However, if the user adds/removes plugins afterwards it's gonna be strange...

        //    //SimpleJSON.JSONNode pluginNode = pluginManager.GetJSON(true, true)["plugins"];
        //    //if (pluginNode == null)
        //    //    SuperController.LogMessage($"{nameof(pluginNode)} == null");

        //    //SimpleJSON.JSONArray pluginNodeArr = pluginNode.AsArray;
        //    //foreach (SimpleJSON.JSONNode node in pluginNodeArr.Childs)
        //    //{
        //    //    SuperController.LogMessage(node.Value);
        //    //}

        //    //return;
        //    // Browsing can already be done, check TEST button already. So we know what you have selected as user, what to trigger on later etc...
            
        //    //bool prevMergeRestore = pluginManager.mergeRestore = true;
        //    //bool isPresetRestore = pluginManager.isPresetRestore = true;
        //    //try
        //    //{
        //    //    pluginManager.mergeRestore = true;
        //    //    pluginManager.isPresetRestore = true;

        //    //    // JSONClass pluginJsonClass = null;
        //    //    // pluginManager.LateRestoreFromJSON(pluginJsonClass); // Would this work!?
        //    //    // Then, for all newly added plugins, simply assign them a new pluginLabelJSON.val to determine when to remove them... ?
        //    //}
        //    //finally
        //    //{
        //    //    pluginManager.isPresetRestore = isPresetRestore;
        //    //    pluginManager.mergeRestore = prevMergeRestore;
        //    //}

        //    //pluginManager.RemoveAllPlugins();

        //    //MVRPlugin plugin = pluginManager.CreatePlugin();
        //    //SuperController.LogMessage(plugin.uid); // plugin#0

        //    //plugin.pluginURLJSON.SetFilePath("Custom/Scripts/Emissive/EmissivePlus.cslist"); // Works
        //    //// plugin.pluginURLJSON.SetFilePath("Blazedust.EmissiveClothingPlus.1:/Custom/Scripts/Emissive/EmissivePlus.cslist"); // Works

        //    //// SuperController.LogMessage(plugin.uid); // plugin#0 (same after SetFilePath, OK!)

        //    //foreach (MVRScriptController scriptController in plugin.scriptControllers)
        //    //{
        //    //    // MVRScript : JSONStorable -> 
        //    //    MVRScript subScript = scriptController.script;
        //    //    SuperController.LogMessage($"subScript.pluginLabelJSON.val: {subScript.pluginLabelJSON.val}"); // empty... user label probably (not verified)
        //    //    SuperController.LogMessage($"subScript.storeId: {subScript.storeId}"); // plugin#0_EmissiveClothing.EmissiveClothingPlus | plugin#0_EmissiveClothing.EmissiveExtraEffects
        //    //}
        //    // plugin.scriptControllers.script
        //}

    }

    public class PluginQueueItem
    {
        public Atom RootAtom;
        public bool Mode;
        public AutoLoadTriggerInfo TriggerInfo;

        public PluginQueueItem(Atom a, bool m, AutoLoadTriggerInfo t)
        {
            RootAtom = a;
            Mode = m;
            TriggerInfo = t;
        }
    }

    public class PluginPresetInfo
    {
        public bool IsValid;
        public string PluginPreset;
        public List<string> PluginURLs;
    }

    public class ScriptInfo
    {
        public string scriptUid;
        public string scriptLabel { 
            get { return Script.pluginLabelJSON.val; } 
            set { Script.pluginLabelJSON.val = value; } 
        }
        public MVRScript Script;
        
        public void Dispose()
        {
            Script = null;
        }
    }


    public class PluginInfo
    {
        public int pluginSequenceNr;
        public string pluginUid;
        public string pluginURL;
        public List<ScriptInfo> Scripts;

        public void Dispose()
        {
            if (Scripts != null)
            {
                foreach(ScriptInfo scriptInfo in Scripts)
                {
                    scriptInfo.Dispose();
                }
                Scripts.Clear();
                Scripts = null;
            }
        }
    }

    public class PluginManagerInfo
    {
        public Atom RootAtom;
        public MVRPluginManager PluginManager;
        public List<PluginInfo> Plugins;

        public PluginManagerInfo(Atom rootAtom)
        {
            RootAtom = rootAtom;
            Plugins = new List<PluginInfo>();
        }

        // Helper to init the UI. Seem like everything is more stable once this has been initialized.
        public bool EnsureUIInit(bool verbose = true)
        {
            if (PluginManager.pluginListPanel != null)
                return true;

            // Need to init pluginListPanel, must open the tab
            if (verbose)
                SuperController.LogMessage("CUAM: PluginManager.InitUI");

            var currentController = SuperController.singleton.GetSelectedController();

            // try init, make sure the root atom has been selected as it init stuff required for PluginManager.InitUI()
            SuperController.singleton.SelectController(RootAtom.mainController);
            PluginManager.InitUI();

            // restore user selection
            SuperController.singleton.SelectController(currentController);
            if (PluginManager.pluginListPanel == null)
            {
                if (verbose)
                    SuperController.LogMessage("CUAM: PluginManager.InitUI Failed");
                return false;
            }

            return true;
        }

        public bool ClearPluginUid(string pluginUid, bool verbose = true)
        {
            if (!EnsureUIInit(verbose))
                return false;

            // Otherwise, for the jSONStorableUrl, it would be SetFilePath(..) ??
            // mvrp.pluginURLJSON
            // MVRPlugin.pluginURLJSON.SetFilePath(..) <--- This would actually REPLACE... hmm! But not load anything.
            // MVRScript : JSONStorable
            // MVRScript.RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
            // MVRScript.LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
            // Condition: IF trying to replace 1, and only 1, plugin from a plugin preset, then user MVRScript RestoreFromJSON(...) function instead. Hmm!

            foreach (Transform t in PluginManager.pluginListPanel.GetComponentsInChildren<Transform>(true))
            {
                foreach (MVRPluginUI pu in t.GetComponentsInChildren<MVRPluginUI>(true).ToArray())
                {
                    if (pu.urlText != null && !string.IsNullOrEmpty(pu.urlText.text) && pu.uidText.text == pluginUid)
                    {
                        if (pu.clearButton != null && pu.clearButton.onClick != null)
                        {
                            if (verbose)
                                SuperController.LogMessage("CUAM: Clearing plugin " + pu.uidText.text + " => " + pu.urlText.text);
                            // pu.enabled = true;
                            // pu.clearButton.OnSubmit(null);
                            pu.clearButton.onClick.Invoke();
                        }
                    }
                }
            }
            return true;
        }

        public void Dispose()
        {
            RootAtom = null;
            PluginManager = null;
            if (Plugins != null)
            {
                foreach (PluginInfo pluginInfo in Plugins)
                {
                    pluginInfo.Dispose();
                }
                Plugins.Clear();
                Plugins = null;
            }
        }

        /// <summary>
        /// The pluginUid data goes STALE when plugin presets are loaded! Might even go stale if adding or removing too!
        /// Read this again before analyzing plugin data.
        /// </summary>
        public void ReadCurrentPluginStatus(bool verbose = true)
        {
            if (RootAtom == null)
            {
                return;
            }

            MVRPluginManager pluginManager = RootAtom.GetStorableByID("PluginManager") as MVRPluginManager;
            PluginManager = pluginManager;
            Plugins.Clear();

            if (pluginManager == null)
            {
                SuperController.LogMessage($"{nameof(pluginManager)} == null");
                return;
            }

            EnsureUIInit(verbose);

            // list existing plugins
            JSONClass json = pluginManager.GetJSON(true, true, true);
            if (json == null)
                SuperController.LogMessage($"{nameof(json)} == null");

            JSONNode jsonNodePlugins = json["plugins"];
            JSONClass jsonClassPlugins = jsonNodePlugins.AsObject;

            List<string> scriptKeys = new List<string>();
            foreach (string storeableId in RootAtom.GetStorableIDs())
            {
                if (storeableId.StartsWith("plugin#"))
                {
                    scriptKeys.Add(storeableId);
                }
            }

            foreach (string key in jsonClassPlugins.Keys)
            {
                PluginInfo pluginInfo = new PluginInfo();
                pluginInfo.pluginUid = key;
                pluginInfo.pluginURL = jsonClassPlugins[key].Value;
                pluginInfo.pluginSequenceNr = GetPluginSequence(key);
                if (pluginInfo.pluginSequenceNr == -1)
                    pluginInfo.pluginSequenceNr = Plugins.Count;
                pluginInfo.Scripts = new List<ScriptInfo>();
                Plugins.Add(pluginInfo);

                // find script storables starting with the plugin key
                foreach (string scriptKey in scriptKeys)
                {
                    if (scriptKey.StartsWith(key + "_"))
                    {
                        ScriptInfo scriptInfo = new ScriptInfo();
                        scriptInfo.scriptUid = scriptKey;
                        scriptInfo.Script = RootAtom.GetStorableByID(scriptKey) as MVRScript;
                        pluginInfo.Scripts.Add(scriptInfo);
                    }
                }
            }

            Plugins.Sort((a, b) => {
                if (a.pluginSequenceNr < b.pluginSequenceNr)
                    return -1;
                if (a.pluginSequenceNr > b.pluginSequenceNr)
                    return 1;
                return 0;
            });
        }

        public int GetPluginSequence(string pluginUid)
        {
            int index = pluginUid.IndexOf("#");
            if (index >= 0)
                pluginUid = pluginUid.Substring(index + 1);

            index = pluginUid.IndexOf("_");
            if (index >= 0)
                pluginUid = pluginUid.Substring(0, index);

            int result = -1;
            if (int.TryParse(pluginUid, out result))
                return result;

            return -1;
        }

        public void Log()
        {
            foreach(PluginInfo pluginInfo in Plugins)
            {
                SuperController.LogMessage(pluginInfo.pluginUid + " => " + pluginInfo.pluginURL);
                foreach(ScriptInfo scriptInfo in pluginInfo.Scripts)
                {
                    SuperController.LogMessage(scriptInfo.scriptUid + " => Label: " + scriptInfo.scriptLabel);
                }
            }
        }
    }
}