using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{

	/// <summary>
    /// This represents multiple CustomUnityAssets in one preset.
    /// Once all atoms have spawned, we know the unique created uids. 
    /// These unique ids will be used to update any referenced atom uids in any scripts in the presets.
    /// This is only useful when the CustomUnityAsset contains a script!
    /// </summary>
	public class PresetCuaPackage
	{

		public int expectedCount;
		public int currentCount;

		private string personAtomUid;

		private Queue<KeyValuePair<JSONNode, Action<JSONNode>>> presetReadyActions;
		private List<JSONNode> presets;
		private Dictionary<string, string> uids;
		private bool updateScriptAtomUids;

		public PresetCuaPackage(int expectedCount, bool updateScriptAtomUids, string rootAtomUid)
		{
			this.expectedCount = expectedCount;
			this.updateScriptAtomUids = updateScriptAtomUids;
			currentCount = 0;
			presetReadyActions = new Queue<KeyValuePair<JSONNode, Action<JSONNode>>>();
			presets = new List<JSONNode>();
			uids = new Dictionary<string, string>();
			personAtomUid = rootAtomUid;
		}

		public bool Update()
		{
			if (currentCount >= expectedCount)
			{
				// process ReadyActions (if any was added, can be 0 if preset contains invalid atom types not configured for the plugin to handle!)
				if (presetReadyActions.Count > 0)
				{
					KeyValuePair<JSONNode, Action<JSONNode>> p = presetReadyActions.Dequeue();
					p.Value(p.Key); // THIS INVOKES the actual creation and update of the atom in the CustomUnityAssetManger.cs once all contained asset atoms have been created.
				}
				return true;
			}
			return false;
		}

		public bool IsCompleted()
		{
			return (currentCount >= expectedCount && presetReadyActions.Count == 0);
		}

		public void Add(string originalAtomUid, Atom apAtom, JSONNode preset, Action<JSONNode> presetProcessed)
		{
			currentCount += 1;
			presets.Add(preset);
			presetReadyActions.Enqueue(new KeyValuePair<JSONNode, Action<JSONNode>>(preset, presetProcessed));
			uids[originalAtomUid] = apAtom.uid;
			CheckReachedExpectedCount();
		}

		/// <summary>
        /// In case of unexpected data in json file or plugin not configured to handle a certain type of atom, it will skip one atom.
        /// This is how we track it.
        /// </summary>
		public void OneNotFound()
		{
			currentCount += 1;
			CheckReachedExpectedCount();
		}

		/// <summary>
        /// One expected count is reached all presets to load is ready and uids can be updated.
        /// </summary>
		private void CheckReachedExpectedCount()
        {
			if (currentCount >= expectedCount)
			{
				//foreach(KeyValuePair<string, string> p in uids)
				//{
				//	SuperController.LogMessage(string.Format("{0} -> {1}", p.Key, p.Value));
				//}

				// Check if renaming script atom uids are requested.
				if (updateScriptAtomUids) // && uids.Any((x) => { return x.Key != x.Value; }))
				{
					foreach (JSONNode p in presets)
					{
						UpdateUidsInPreset(p, uids, personAtomUid);
					}
				}
			}
		}

		private static Dictionary<string, AtomRenamingStorableTargetItem[]> atomRenameTargets = null;

		/// <summary>
		/// This is based on storing CustomUnityAssets and updating scripts iniside CustomUnityAsset JSON structures.
		/// Code here assumes you have knowledge over preset JSON file structure in VaM. 
		/// This structure may update in future updates of VAM and may break this feature in this plugin.
		/// </summary>
		/// <param name="preset"></param>
		/// <param name="uids"></param>
		private static void UpdateUidsInPreset(JSONNode preset, Dictionary<string, string> uids, string personAtomUid)
        {
			try
            {
				if (preset["atoms"] != null)
				{

					// initialize renaming targets!
					if (atomRenameTargets == null)
					{	
						atomRenameTargets = new Dictionary<string, AtomRenamingStorableTargetItem[]>();
						atomRenameTargets.Add("*", new AtomRenamingStorableTargetItem[] {
							new AtomRenamingStorableTargetItem("*", "control", "linkTo", true)
						});
						atomRenameTargets.Add("CycleForce", new AtomRenamingStorableTargetItem[] { 
							new AtomRenamingStorableTargetItem("SyncForce", "ForceProducer", "receiver", true),
							new AtomRenamingStorableTargetItem("SyncForce", "ForceProducer", "syncTo", false)
						});
						atomRenameTargets.Add("SyncForce", new AtomRenamingStorableTargetItem[] {
							new AtomRenamingStorableTargetItem("SyncForce", "ForceProducer", "receiver", true)
						});
						atomRenameTargets.Add("GrabPoint", new AtomRenamingStorableTargetItem[] {
							new AtomRenamingStorableTargetItem("GrabPoint", "GrabPoint", "controlled", true)
						});
						
					}

					JSONArray jnAtoms = preset["atoms"].AsArray;
					if (jnAtoms != null && jnAtoms.Count > 0)
                    {
						HashSet<string> origUids = new HashSet<string>(uids.Keys); // easier lookup

						foreach (JSONNode jnAtom in jnAtoms.Childs)
                        {
							string atomType = jnAtom["type"]?.Value;
							AtomRenamingStorableTargetItem[] renameTargets = null;
							if (!atomRenameTargets.TryGetValue((atomType != null ? atomType : ""), out renameTargets)) { renameTargets = null; }

							if (jnAtom["storables"] != null)
                            {
								JSONArray stores = jnAtom["storables"].AsArray;

								string[] pluginKeys = null;

								foreach (JSONNode store in stores)
                                {
									if (store["id"] == null)
                                    {
										continue;
                                    }

									if (store["id"].Value == "PluginManager")
									{
										// Found PluginManager!
										pluginKeys = store["plugins"].AsObject.Keys.ToArray();
									}

									if (store["id"].Value == "control")
                                    {
										// special control rename target, always 1
										RenameTargets(store, atomRenameTargets["*"][0], uids, personAtomUid);
									}

									if (renameTargets != null)
                                    {
										AtomRenamingStorableTargetItem renameTarget = renameTargets.FirstOrDefault((x) => { return x.storableId == store["id"].Value; });
										if (renameTarget != null)
                                        {
											RenameTargets(store, renameTarget, uids, personAtomUid);
										}
									}

								}

								if (pluginKeys != null)
                                {
									// SuperController.LogMessage("PluginManager found with scripts: " + string.Join(", ", pluginKeys));

									foreach (JSONNode store in stores)
									{
										if (store["id"] != null && pluginKeys.Any((x) => { return store["id"].Value.StartsWith(x + "_"); }))
										{
											// SuperController.LogMessage("Found a plugin!");
											// This store node needs to be checked with updated string values!
											JSONClass jcScript = store.AsObject;
											string[] pluginValueKeys = jcScript.Keys.ToArray();

											// SuperController.LogMessage(string.Format("Plugin {0} contains following keys: {1}", store["id"].Value, string.Join(", ", pluginValueKeys)));
											foreach(string valueKey in pluginValueKeys)
                                            {
												if (origUids.Contains(store[valueKey].Value))
                                                {
													// rename time!
													// SuperController.LogMessage(string.Format("Updating key {0} value '{1}' to '{2}'", valueKey, store[valueKey].Value, uids[store[valueKey].Value]));
													store[valueKey].Value = uids[store[valueKey].Value];
												}
                                            }
										}
									}
								}
							}
                        }
                    }
				}
			} 
			catch (Exception e)
            {
				SuperController.LogError("Exception during UpdateUidsInPreset. You might want to disable 'Update AtomUids In Saved Scripts' trying to load this preset.");
				SuperController.LogError(e.ToString());
			}
			
			// SuperController.LogMessage(preset["atoms"].AsArray[0]["type"]);
		}

		private static void RenameTargets(JSONNode store, AtomRenamingStorableTargetItem renameTarget, Dictionary<string, string> uids, string personAtomUid)
        {
			// SuperController.LogMessage("Renaming store " + store["id"].Value + " for key " + renameTarget.key + " for player: " + renameTarget.targetPlayer);
			if (store[renameTarget.key] != null)
			{
				string value = store[renameTarget.key].Value;
				if (value != null && value != "")
                {
					// value is in format "person:link" (for persons) and "atomuid:store" (for other atoms)
					int index = value.LastIndexOf(":");
					if (index > 0)
                    {
						if (renameTarget.targetPlayer)
						{
							value = personAtomUid + value.Substring(index); // value.Substring(index) includes ":"
						} 
						else
                        {
							string atomUid = value.Substring(0, index - 1);
							if (uids.ContainsKey(atomUid))
                            {
								value = uids[atomUid] + value.Substring(index); // value.Substring(index) includes ":"
							}
						}
					} 
					else
                    {
						if (renameTarget.targetPlayer)
                        {
							value = personAtomUid;
                        } 
						else
                        {
							if (uids.ContainsKey(value))
							{
								value = uids[value];
							}
						}
                    }


					if (store[renameTarget.key].Value != value)
                    {
						// SuperController.LogMessage("Updating " + renameTarget.key + " from " + store[renameTarget.key].Value + " to " + value);
						store[renameTarget.key].Value = value;
					}

				}

			}
        }

	}

	/// <summary>
    /// Represents a uid storable id/key - value to update to the person's id this is applied on!
    /// </summary>
	class AtomRenamingStorableTargetItem
    {
		public string atomType;
		public string storableId;
		public string key;
		public bool targetPlayer;

		public AtomRenamingStorableTargetItem(string atomType, string storableId, string key, bool targetPlayer)
        {
			this.atomType = atomType;
			this.storableId = storableId;
			this.key = key;
			this.targetPlayer = targetPlayer;
		}

	}

}