using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;

namespace Blazedust.CUAM
{

    /// <summary>
    /// Manager to track created controls to easily remove them.
    /// </summary>
    public class UIHelper
    {

        public List<UIDynamic> controls;
        public MVRScript core;

        public UIHelper(MVRScript script)
        {
            controls = new List<UIDynamic>();
            core = script;
        }

        public UIDynamicTextField CreateLabel(string label, bool rhs, int height=40)
        {
            JSONStorableString jsonLabel = new JSONStorableString(label, label);
            UIDynamicTextField labelField = CreateTextField(jsonLabel, rhs);
            UIHelper.SetTextFieldHeight(labelField, height);

            return labelField;
        }

        public UIDynamicTextField CreateStringTextbox(JSONStorableString json, string label, bool rhs)
        {
            if (label != null && label != "")
            {
                JSONStorableString jsonLabel = new JSONStorableString(json.name + "_label", label);
                UIDynamicTextField labelField = CreateTextField(jsonLabel, rhs);
                UIHelper.SetTextFieldHeight(labelField, 40);
            }

            UIDynamicTextField textField = CreateTextField(json, rhs);
            UIHelper.SetTextFieldHeight(textField, 60);
            InputField input = textField.gameObject.AddComponent<InputField>();
            input.textComponent = textField.UItext;
            input.textComponent.fontSize = 32;

            Transform t = textField.GetComponent<Transform>();
            t.Translate(0f, 100f, 0f);

            textField.backgroundColor = Color.white;
            json.inputField = input;

            input.onValueChanged.AddListener((string value) => {
                json.val = value;
            });

            return textField;
        }

        public UIDynamic CreateSpacer(bool rightSide, int height)
        {
            UIDynamic spacer = core.CreateSpacer(rightSide);
            controls.Add(spacer);
            spacer.height = height;
            return spacer;
        }

        public UIDynamicButton CreateButton(string label, bool rightSide = false)
        {
            UIDynamicButton ctrl = core.CreateButton(label, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public UIDynamicColorPicker CreateColorPicker(JSONStorableColor jsc, bool rightSide = false)
        {
            UIDynamicColorPicker ctrl = core.CreateColorPicker(jsc, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public UIDynamicTextField CreateTextField(JSONStorableString jss, bool rightSide = false)
        {
            UIDynamicTextField ctrl = core.CreateTextField(jss, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public UIDynamicPopup CreatePopup(JSONStorableStringChooser jsc, bool rightSide = false)
        {
            UIDynamicPopup ctrl = core.CreatePopup(jsc, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public UIDynamicPopup CreateScrollablePopup(JSONStorableStringChooser jsc, bool rightSide = false)
        {
            UIDynamicPopup ctrl = core.CreateScrollablePopup(jsc, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public UIDynamicPopup CreateFilterablePopup(JSONStorableStringChooser jsc, bool rightSide = false)
        {
            UIDynamicPopup ctrl = core.CreateFilterablePopup(jsc, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }
        public void OpenUp(UIDynamicPopup popup)
        {
            popup.popup.popupPanel.localPosition = new Vector3(popup.popup.popupPanel.localPosition.x, popup.popupPanelHeight + 55f, popup.popup.popupPanel.localPosition.z);
        }

        public UIDynamicToggle CreateToggle(JSONStorableBool jsb, bool rightSide = false)
        {
            UIDynamicToggle ctrl = core.CreateToggle(jsb, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public UIDynamicSlider CreateSlider(JSONStorableFloat jsf, bool rightSide = false)
        {
            UIDynamicSlider ctrl = core.CreateSlider(jsf, rightSide);
            controls.Add(ctrl);
            return ctrl;
        }

        public void Remove(UIDynamic ctrl)
        {
            controls.Remove(ctrl);
            if (ctrl is UIDynamicToggle)
            {
                core.RemoveToggle((UIDynamicToggle)ctrl);
            } 
            else if (ctrl is UIDynamicSlider)
            {
                core.RemoveSlider((UIDynamicSlider)ctrl);
            }
            else if (ctrl is UIDynamicColorPicker)
            {
                core.RemoveColorPicker((UIDynamicColorPicker)ctrl);
            }
            else if (ctrl is UIDynamicTextField)
            {
                core.RemoveTextField((UIDynamicTextField)ctrl);
            }
            else if (ctrl is UIDynamicPopup)
            {
                core.RemovePopup((UIDynamicPopup)ctrl);
            }
            else if (ctrl is UIDynamicButton)
            {
                core.RemoveButton((UIDynamicButton)ctrl);
            }
            else if (ctrl is UIDynamic)
            {
                core.RemoveSpacer((UIDynamic)ctrl);
            }
        }

        public void RemoveAllControls()
        {
            // SuperController.LogMessage("Removing " + controls.Count.ToString() + " controls.");
            for(int i = controls.Count - 1; i >= 0; i--)
            {
                Remove(controls[i]);
            }
        }

        public void SetEnabled(bool enabled)
        {
            for (int i = controls.Count - 1; i >= 0; i--)
            {
                SetEnabled(controls[i], enabled);
            }
        }

        /// <summary>
        /// This moves the control out-of-sight rather than anything else.
        /// </summary>
        public void SetEnabled(UIDynamic control, bool enabled)
        {
            bool updateRect = false;
            LayoutElement component = control.GetComponent<LayoutElement>();
            if (component != null)
            {
                updateRect = (component.enabled != enabled);
                component.enabled = enabled;
            }
            RectTransform component2 = control.GetComponent<RectTransform>();
            if (component2 != null)
            {
                Vector3 t = component2.localPosition;
                if (enabled)
                {
                    t.x -= 1200f;
                }
                else
                {
                    t.x += 1200f;
                }
                component2.localPosition = t;
            }
        }

        public void Destroy()
        {
            RemoveAllControls();
        }

        public static void SetTextFieldHeight(UIDynamicTextField textField, int height)
        {
            LayoutElement component = textField.GetComponent<LayoutElement>();
            if (component != null)
            {
                component.minHeight = height;
                component.preferredHeight = height;
            }
            textField.height = height;
        }

    }
}